#version 400 core

uniform sampler2D shipsels;
uniform sampler2D springs;
uniform sampler2D data;
uniform float delta;
uniform float time;
uniform float wh = 9;
uniform float ww = 6;
uniform vec4 tool = vec4(0);

layout(location = 0) out vec4 shipsel;
layout(location = 1) out vec4 outdata;
layout(location = 2) out vec4 spring;

uniform float floodspeed = 1;

const float WATER = 1080.;
const float AIR = 1.225;
uniform float K = 150000.;
uniform float strength = 5000.;
uniform float dragmul = 0.005;
uniform float seafloor = -200;
uniform float waterweight = 1;
uniform float buoyancy = 1;

vec2    getJointForces(ivec2 dir, ivec2 dims, vec4 actualPos){
    vec2 force = vec2(0);
    int i = 0;
    for (i = 1; i <= int(sqrt(K * 0.01)) ; ++i)
    {
        ivec2 displacement = i * dir;
        ivec2 coord = ivec2(gl_FragCoord.xy) + displacement;
        if (coord.x < 0 || coord.y < 0 || coord.x >= dims.x || coord.y >= dims.y)
            break;
        vec4 shipselData = texelFetch(data, coord, 0);
        if (shipselData == vec4(0))
           return force;
        vec4 shipselPos = texelFetch(shipsels, coord, 0);
        vec2 x = actualPos.xy - shipselPos.xy;
        if (x == vec2(0))
             continue;
        float xl = length(x);
        force += -K / (i * 0.33) * (x / xl) * (xl - length(displacement)) - (actualPos.zw - shipselPos.zw) * sqrt(K * 0.01);
    }
    return force;
}

float getFluidDensity(vec2 pos)
{
    float x = pos.x / ww;
    float y = (sin(x * 0.1f + time) * 0.5f + sin(x * 0.3f - time * 1.1f) * 0.3f) * wh;
    if (pos.y > y)
        return (AIR);
    else
        return (WATER);
}

float flood(vec4 curdata, vec4 curpos,  ivec2 res, float density){
    float newflood = curdata.z;
    if (density == WATER)
    {
    vec4 up, down, right, left, downleft, downright;
    vec4 a2, b2, c2, d2;
    up = texelFetch(data, ivec2(gl_FragCoord.xy) + ivec2(0, 1), 0);
    down = texelFetch(data, ivec2(gl_FragCoord.xy) + ivec2(0, -1), 0);
    right = texelFetch(data, ivec2(gl_FragCoord.xy) + ivec2(1, 0), 0);
    left = texelFetch(data, ivec2(gl_FragCoord.xy) + ivec2(-1, 0), 0);
    downleft = texelFetch(data, ivec2(gl_FragCoord.xy) + ivec2(-1, -1), 0);
    downright = texelFetch(data, ivec2(gl_FragCoord.xy) + ivec2(1, -1), 0);

    if (up.x == 0 || down.x == 0 ||right.x == 0 || left.x == 0)
        newflood += (density == AIR? -1 : 1) * distance(1, curdata.z) * delta * floodspeed;
    if (gl_FragCoord.y < res.y && (up.w) != 1 && up.z > 0){
        newflood += ((clamp(up.z - curdata.z, -0.2, 1.) + 0.15) * floodspeed) * delta;
    }
    if (gl_FragCoord.y > 0 && (down.w) != 1 && down.z > 0 && (downright.z > 0.9 || downright.w == 1) && (downleft.z > 0.9 || downleft.w == 1)){
        newflood += ((clamp(down.z - curdata.z, -0.2, 1.) + 0.15) * floodspeed) * delta * 0.1;
    }
    if (gl_FragCoord.x < res.x && (right.w) != 1 && right.z > 0 && (down.z > 0.9 || down.w == 1 || downright.w == 1)){
        newflood += ((clamp(right.z - curdata.z, -0.2, 1.) + 0.15) * floodspeed) * delta;
    }
    if (gl_FragCoord.x > 0 && (left.w) != 1 && left.z > 0 && (down.z > 0.9 || down.w == 1 || downleft.w == 1)){
        newflood += ((clamp(left.z - curdata.z, -0.2, 1.) + 0.15) * floodspeed) * delta;
    }
    }else
    {
        newflood -= delta * floodspeed * 0.1;
    }
    return clamp(newflood, 0, 1);
}

void    main() {
    outdata = texelFetch(data, ivec2(gl_FragCoord.xy), 0);
    if (outdata == vec4(0))
        return;
    shipsel = texelFetch(shipsels, ivec2(gl_FragCoord.xy), 0);
    spring = texelFetch(springs, ivec2(gl_FragCoord.xy), 0);
    ivec2 res = textureSize(shipsels, 0);

    float density = getFluidDensity(shipsel.xy);
    outdata.z = flood(outdata, shipsel, res, density);
    vec2 fu = getJointForces(ivec2(0, 1), res, shipsel);
    vec2 fd = getJointForces(ivec2(0, -1), res, shipsel);
    vec2 fr = getJointForces(ivec2(1, 0), res, shipsel);
    vec2 fl = getJointForces(ivec2(-1, 0), res, shipsel);
    vec2 fur = getJointForces(ivec2(1, 1), res, shipsel);
    vec2 fdl = getJointForces(ivec2(-1, -1), res, shipsel);
    vec2 fdr = getJointForces(ivec2(1, -1), res, shipsel);
    vec2 flu = getJointForces(ivec2(-1, 1), res, shipsel);
    vec2 force = (fu + fd + fr + fl + fur + fdl + fdr + flu) / 8;
    float strain = sqrt(-min(min(min(dot(fu, fd), dot(fr, fl)), dot(fur, fdl)), dot(flu, fdr)));
    vec2 drag = vec2(0);
    // calculate drag relative to surface
    vec2 norm = normalize(shipsel.zw);
    if (shipsel.zw != vec2(0) && (fu == vec2(0) || fd == vec2(0) || flu == vec2(0) || fr == vec2(0) || fl == vec2(0) || fur == vec2(0) || fdl == vec2(0) || fdr == vec2(0)))
    {
        float a = dot(norm, vec2(1, 0));
        float b = dot(norm, vec2(0, 1));
        if (a >= 0.5){
            if (b >= 0){
                if (fr == vec2(0) || fur == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }else{
                if (fr == vec2(0) || fdr == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }
        }else if (a >= 0){
            if (b >= 0.5){
                if (fu == vec2(0) || fur == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }else{
                if (fd == vec2(0) || fdr == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }
        }else if (a >= -0.5){
            if (b >= 0.5){
                if (fu == vec2(0) || flu == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }else{
                if (fd == vec2(0) || fdl == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }
        }else{
            if (b >= 0){
                if (fl == vec2(0) || flu == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }else{
                if (fl == vec2(0) || fdl == vec2(0))
                    drag = -norm * density * dragmul * dot(shipsel.zw, shipsel.zw);
            }
        }
    }
    if (shipsel.zw != vec2(0) && drag == vec2(0))
        drag = -norm * density * dragmul * 0.1 * dot(shipsel.zw, shipsel.zw);

    float mass;
    mass = mix(outdata.x, mix(AIR, WATER * waterweight, clamp(outdata.z, 0, 1)), 0.95) / (waterweight + buoyancy * ( 1 - outdata.z));
    vec4 lastShipsel = shipsel;
    vec2 f = vec2(0);
    if (tool != vec4(0))
    {
    int curtool = floatBitsToInt(tool.w);
    if (curtool == 1)
    {
        if (tool.z > 0);
            f = normalize(tool.xy - shipsel.xy) * tool.z * 10;
    } else if (curtool == 2)
    {
        if (tool.z > 0 && distance(shipsel.xy, tool.xy) < tool.z)
        {
            outdata = vec4(0);
            return;
        }

        } else if (curtool == 3)
        {
            if (tool.z > 0 && distance(shipsel.xy, tool.xy) < tool.z)
            {
                outdata.z = clamp(outdata.z + delta * tool.z, 0, 1);
                return;
            }
        }else if (curtool == 4)
        {
            if (tool.z > 0 && distance(shipsel.xy, tool.xy) < tool.z)
            {
                outdata.z = clamp(outdata.z - delta * tool.z, 0, 1);
                return;
            }

        }
    }
    if (shipsel.y < seafloor)
    {
        shipsel.y = seafloor;
        shipsel.zw *= -0.5;
    }
    shipsel.zw += delta * (drag + vec2(0, -9.81 * waterweight) + f +((vec2(0, (density * 9.81 )) + force) / mass));
    shipsel.xy += shipsel.zw * delta;
    if (strain > outdata.y * strength)
    {
        outdata = vec4(0);
        return;
    }
    /*
    //mathematical integration of the formula broken (maybe)
    vec2 result = delta * (mass * (drag +  vec2(0, -9.81)) + density + 9.81 + force) / mass;
    shipsel.xy = shipsel.xy + delta * shipsel.zw + result * 0.5 * delta;
    shipsel.zw = shipsel.zw + result;
    */
}
